#!/usr/bin/env python

import asyncio

from ruia import Request
from ruia.exceptions import InvalidRequestMethod

sem = asyncio.Semaphore(3)


async def hello(response):
    return 'hello ruia'


def hi(response):
    yield 'hi ruia'


async def retry_func(request):
    request.request_config['TIMEOUT'] = 10


params = {
    "name": "ruia"
}


async def valid_response(response):
    return response


async def make_get_request(sem, callback=None):
    request_config = {
        'RETRIES': 3,
        'DELAY': 1,
        'TIMEOUT': 0.1,
        'VALID': valid_response,
        'RETRY_FUNC': retry_func
    }
    request = Request('https://httpbin.org/get',
                      method='GET',
                      metadata={'hello': 'ruia'},
                      request_config=request_config,
                      params=params,
                      callback=callback)
    return await request.fetch_callback(sem)


async def make_post_request(sem, callback):
    headers = {
        'Content-Type': 'application/json'
    }
    request = Request('https://httpbin.org/post',
                      method='POST',
                      headers=headers,
                      data=params,
                      callback=callback)
    return await request.fetch_callback(sem)


def test_request_config():
    assert str(Request('https://httpbin.org/')) == '<GET https://httpbin.org/>'
    _, response = asyncio.get_event_loop().run_until_complete(make_get_request(sem=sem, callback=hello))
    assert response.callback_result == 'hello ruia'
    assert response.metadata == {'hello': 'ruia'}
    json_result = asyncio.get_event_loop().run_until_complete(response.json())
    assert json_result['args']['name'] == "ruia"

    _, response = asyncio.get_event_loop().run_until_complete(make_post_request(sem=sem, callback=None))
    json_result = asyncio.get_event_loop().run_until_complete(response.json())
    assert json_result['data'] == "name=ruia"


def test_method_error_request():
    try:
        request = Request('https://httpbin.org/', method='PUT')
        response = asyncio.get_event_loop().run_until_complete(request.fetch())
        assert response.html == ''
    except Exception as e:
        assert isinstance(e, InvalidRequestMethod)


def test_sem_error_request():
    _, response = asyncio.get_event_loop().run_until_complete(make_get_request(sem=None, callback=None))
    assert response == None


def test_retry_request():
    request = Request('http://httpbin.org/404')
    _, response = asyncio.get_event_loop().run_until_complete(request.fetch_callback(sem=sem))
    assert response.url == 'http://httpbin.org/404'


def test_timeout_request():
    async def timeout_request(sem):
        request_config = {
            'RETRIES': 1,
            'DELAY': 1,
            'TIMEOUT': 0.1,
        }
        request = Request('https://httpbin.org/get',
                          method='GET',
                          metadata={'hello': 'ruia'},
                          encoding='utf-8',
                          request_config=request_config,
                          params=params,
                          callback=hi)
        return await request.fetch_callback(sem)

    _, response = asyncio.get_event_loop().run_until_complete(timeout_request(sem=sem))
    assert response.url == 'https://httpbin.org/get'



import ruia

from lxml import etree

HTML = '''
<body>
<div class="title" href="/">Ruia Documentation</div>
<ul>
    <li class="tag" href="./easy.html">easy</li>
    <li class="tag" href="./fast.html">fast</li>
    <li class="tag" href="./powerful.html">powerful</li>
</ul>
</body>
'''

html = etree.HTML(HTML)


def test_text_field():
    title = ruia.TextField(css_select='.title', default='Untitled')
    assert title.extract(html_etree=html) == 'Ruia Documentation'
    tags = ruia.TextField(css_select='.tag', default='No tag', many=True)
    assert tags.extract(html_etree=html) == ['easy', 'fast', 'powerful']


def test_attr_field():
    title = ruia.AttrField(css_select='.title', attr='href', default='Untitled')
    assert title.extract(html_etree=html) == '/'
    tags = ruia.AttrField(css_select='.tag', attr='href', default='No tag', many=True)
    assert tags.extract(html_etree=html)[0] == './easy.html'


def test_html_field():
    title = ruia.HtmlField(css_select='.title', default='Untitled')
    assert title.extract(html_etree=html) == '<div class="title" href="/">Ruia Documentation</div>\n'
    tags = ruia.HtmlField(css_select='.tag', default='No tag', many=True)
    assert tags.extract(html_etree=html)[1] == '<li class="tag" href="./fast.html">fast</li>\n    '


def test_regex_field():
    title = ruia.RegexField(re_select='<div class="title" href="(.*?)">(.*?)</div>')
    assert title.extract(html=HTML)[0] == '/'
    assert title.extract(html=HTML)[1] == 'Ruia Documentation'
    tags = ruia.RegexField(re_select='<li class="tag" href="(?P<href>.*?)">(?P<text>.*?)</li>', many=True)
    result = tags.extract(html=HTML)
    assert isinstance(result, list)
    assert len(result) == 3
    assert isinstance(result[0], dict)
    assert result[0]['href'] == './easy.html'



#!/usr/bin/env python

from .field import AttrField, BaseField, HtmlField, RegexField, TextField
from .item import Item
from .middleware import Middleware
from .request import Request
from .response import Response
from .spider import Spider

__version__ = '0.6.0'


#!/usr/bin/env python

import os
import re
import sys

from setuptools import find_packages, setup

_version_re = re.compile(r'__version__\s+=\s+(.*)')

PY_VER = sys.version_info

if PY_VER < (3, 6):
    raise RuntimeError("ruia doesn't support Python version prior 3.6")


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*'([\d.abrc]+)'")
    init_py = os.path.join(os.path.dirname(__file__), 'ruia', '__init__.py')
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)


def read(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name)) as f:
        return f.read()


setup(
    name='ruia',
    version=read_version(),
    author='Howie Hu',
    description="Async Python 3.6+ web scraping micro-framework based on asyncio.",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author_email='xiaozizayang@gmail.com',
    python_requires='>=3.6',
    install_requires=['aiohttp>=3.5.4', 'cssselect', 'lxml'],
    url="https://python-ruia.org/",
    packages=find_packages(),
    license='MIT',
    classifiers=[
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: BSD",
        "Operating System :: Microsoft :: Windows",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={
        'Documentation': 'https://docs.python-ruia.org/',
        'Source': 'https://github.com/howie6879/ruia',
    },
    extras_require={
        'uvloop': ['uvloop']
    }
)



#!/usr/bin/env python

import asyncio
import os

from ruia import Item, Middleware, Response, Request, Spider, TextField

html_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'for_spider_testing.html')
with open(html_path, mode='r', encoding='utf-8') as file:
    HTML = file.read()

middleware = Middleware()


async def retry_func(request):
    request.request_config['TIMEOUT'] = 10


@middleware.request
async def print_on_request(spider_ins, request):
    request.headers = {
        'User-Agent': 'ruia ua'
    }


@middleware.response
async def print_on_response(spider_ins, request, response):
    assert isinstance(response.html, str)
    assert request.headers == {
        'User-Agent': 'ruia ua'
    }


class ItemDemo(Item):
    title = TextField(xpath_select='/html/head/title')


class SpiderDemo(Spider):
    start_urls = ['https://httpbin.org/get?p=0']
    request_config = {
        'RETRIES': 3,
        'DELAY': 0,
        'TIMEOUT': 20,

    }
    headers = {
        "User-Agent": "Ruia Spider"
    }

    call_nums = 0

    async def parse(self, response):
        yield Request(
            url=response.url,
            callback=self.parse_item,
            headers=self.headers,
            request_config=self.request_config,
            **self.kwargs
        )

    async def parse_item(self, response):
        pages = [f'https://httpbin.org/get?p={i}' for i in range(1, 2)]
        async for resp in self.multiple_request(pages):
            yield self.parse_next(response=resp, any_param='hello')

    async def parse_next(self, response, any_param):
        assert any_param == 'hello'
        yield self.request(url='https://httpbin.org/get?p=2',
                           metadata={'any_param': any_param},
                           callback=self.parse_details)

    async def parse_details(self, response):
        item = await ItemDemo.get_item(html=HTML)
        yield item

    async def process_item(self, item: ItemDemo):
        assert item.title == 'for_spider_testing'
        await self.count_nums()

    async def count_nums(self):
        SpiderDemo.call_nums += 1


def test_spider_with_middleware():
    loop = asyncio.new_event_loop()
    SpiderDemo.start(loop=loop, middleware=middleware)
    assert SpiderDemo.call_nums == 1


def test_spider_with_error_middleware():
    error_middleware = Middleware()

    @error_middleware.request
    def error_request(spider_ins, request, response):
        pass

    @error_middleware.response
    async def error_response(spider_ins, request, response):
        raise TypeError('error')

    class SpiderDemo(Spider):
        start_urls = ['https://httpbin.org/get?p=0']

        async def parse(self, response):
            pass

    SpiderDemo.start(middleware=error_middleware)


def test_spider_hook():
    async def after_start_func(spider_ins):
        print("after_start_func")
        spider_ins.result['after_start'] = True
        assert isinstance(spider_ins.result, dict)

    async def before_stop_func(spider_ins):
        print("before_stop_func")
        spider_ins.result['before_stop'] = True

    class SpiderHook(Spider):
        start_urls = ['https://httpbin.org/get?p=0', 'https://httpbin.org/404']
        request_config = {
            'RETRIES': 1,
            'DELAY': 0,
            'TIMEOUT': 10,

        }

        result = {
            'after_start': False,
            'before_stop': False,
            'process_succeed_response': False,
            'process_failed_response': False,
            'process_item': False
        }

        async def parse(self, response):
            item = await ItemDemo.get_item(html=HTML)
            yield item

        async def process_item(self, item):
            self.result['process_item'] = True

        async def process_succeed_response(self, request, response):
            # Hook for response
            self.result['process_succeed_response'] = True

        async def process_failed_response(self, request, response):
            # Hook for response
            self.result['process_failed_response'] = True

    # Test middleware & hook
    loop = asyncio.new_event_loop()
    SpiderHook.start(loop=loop, after_start=after_start_func, before_stop=before_stop_func)

    assert SpiderHook.result['after_start'] == True
    assert SpiderHook.result['before_stop'] == True
    assert SpiderHook.result['process_succeed_response'] == True
    assert SpiderHook.result['process_failed_response'] == True
    assert SpiderHook.result['process_item'] == True


def test_spider_hook_error():
    class SpiderDemo(Spider):
        start_urls = ['https://httpbin.org/get?p=0']

        async def parse(self, response):
            pass

    async def before_stop_func(spider_ins):
        raise TypeError('error')

    loop = asyncio.new_event_loop()
    SpiderDemo.start(loop=loop, before_stop=before_stop_func)


def test_invalid_callback_result():
    class SpiderDemo(Spider):
        start_urls = ['https://httpbin.org/get?p=0']
        result = {
            'process_callback_result': False
        }

        async def parse(self, response):
            yield {}

    async def process_dict_callback_result(spider_ins, callback_result):
        spider_ins.result['process_callback_result'] = True

    class CustomCallbackResultType:

        @classmethod
        def init_spider(cls, spider):
            spider.callback_result_map = spider.callback_result_map or {}
            setattr(spider, 'process_dict_callback_result', process_dict_callback_result)
            spider.callback_result_map.update({'dict': 'process_dict_callback_result'})

    CustomCallbackResultType.init_spider(SpiderDemo)

    loop = asyncio.new_event_loop()
    SpiderDemo.start(loop=loop)
    assert SpiderDemo.result['process_callback_result'] == True


def test_spider_multiple_request_sync():
    result = list()

    class MultipleRequestSpider(Spider):
        start_urls = ['https://httpbin.org']
        concurrency = 3

        async def parse(self, response: Response):
            urls = [f'https://httpbin.org/get?p={page}' for page in range(1, 2)]
            async for response in self.multiple_request(urls, is_gather=True):
                yield self.parse_next(response=response)

        async def parse_next(self, response):
            json_result = await response.json()
            page = json_result['args']['p']
            result.append(int(page))

    MultipleRequestSpider.start()
    assert result == [1]


def test_no_start_url_spider():
    try:
        class NoStartUrlSpider(Spider):
            pass

        NoStartUrlSpider.start()
    except Exception as e:
        assert isinstance(e, ValueError)


def test_callback_error():
    class NoParseSpider(Spider):
        start_urls = ['https://httpbin.org/get']

    NoParseSpider.start()

    class CallbackError(Spider):
        start_urls = ['https://httpbin.org/get']

        async def parse(self, response):
            raise ValueError('error')

    CallbackError.start()


def test_coroutine_callback_error():
    class CoroutineItemErrorSpider(Spider):
        start_urls = ['https://httpbin.org/get']

        async def parse(self, response):
            pages = ['https://httpbin.org/get?p=1']
            async for resp in self.multiple_request(pages):
                yield self.parse_item(response=resp)

        async def parse_item(self, response):
            await ItemDemo.get_item(html=response.html)

    CoroutineItemErrorSpider.start()

    class CoroutineErrorSpider(Spider):
        start_urls = ['https://httpbin.org/get']

        async def parse(self, response):
            pages = ['https://httpbin.org/get?p=1']
            async for resp in self.multiple_request(pages):
                yield self.parse_item(response=resp)

        async def parse_item(self, response):
            raise ValueError("error")

    CoroutineErrorSpider.start()


def test_nothing_marched_spider():
    class NothingMatchedErrorSpider(Spider):
        start_urls = ['https://httpbin.org/get']

        async def parse(self, response):
            await ItemDemo.get_item(html=response.html)

    NothingMatchedErrorSpider.start()


def test_multiple_spider():
    class MultipleSpider(Spider):
        count = 0
        start_urls = ['https://httpbin.org/get?p=0']

        async def parse(self, response):
            MultipleSpider.count += 1

    async def multiple_spider(loop):
        await MultipleSpider.async_start(loop=loop, middleware=[middleware])
        await MultipleSpider.async_start(loop=loop, middleware=middleware)
        return MultipleSpider

    loop = asyncio.new_event_loop()
    spider_ins = loop.run_until_complete(multiple_spider(loop=loop))
    assert spider_ins.count == 2
